/*=========================================================================

Program:   Insight Segmentation & Registration Toolkit
Module:    $RCSfile: CudaDiscreteGaussian.cu,v $
Language:  C/C++ CUDA
Date:      $Date: 2010-09-15 12:44:26 $
Version:   $Revision: 3.1.0 $

Copyright (c) Insight Software Consortium. All rights reserved.
See ITKCopyright.txt or http://www.itk.org/HTML/Copyright.htm for details.

This software is distributed WITHOUT ANY WARRANTY; without even 
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
/**
 * \class CudaDiscreteGaussianImageFilter
 * \brief Blurs an image by separable convolution with discrete gaussian kernels.
 * \author Luis Lourenco - Vision, Robotics, and Image Group (VRI)
 * http://www.inf.ufpr.br/vri/alumni/2011-LuisLourenco/
**/

#include <stdio.h>
#include <stdlib.h>

#include <cuda.h>
#include <cutil.h>

#include "CudaDiscreteGaussian.h"

/// allocate texture variables
texture<float, 1, cudaReadModeElementType> texRef;
texture<float, 1, cudaReadModeElementType> gaussTexRef;



__global__ void calculateGaussianKernel(float *gaussKernel, const float sigma, int halfKernelWidth){

  /// pixel index of this thread
  /// this makes the normal curve
  int i = threadIdx.x - halfKernelWidth;
  extern __shared__ float s_gaussKernel[];
  __shared__ float sum;
  
  /// this kernel must allocate 'kernelWidth' threads
  s_gaussKernel[threadIdx.x] = (__fdividef(1,(sqrtf(2*M_PI*sigma))))*expf((-1)*(__fdividef((i*i),(2*sigma*sigma))));

  __syncthreads();

  /// Thread 0 sum all the gassian kernel array
  // This is not so bad because the array is always short
  if (!threadIdx.x) {
    int th;
    sum = 0;
    for(th = 0; th<blockDim.x; th++) sum += s_gaussKernel[th];
  }

  __syncthreads();

  gaussKernel[threadIdx.x] = s_gaussKernel[threadIdx.x]/sum;

}

float* cuda1DGaussianOperator(dim3 DimGrid, dim3 DimBlock, unsigned int width, float gaussianVariance){

  /// The Gaussian Kernel Width must be odd
  if (width < 1) width = 1;
  if (width%2 == 0) width--;
  short halfWidth = width >> 1;

  int kernelSize = width*sizeof(float);

  float *cudaGaussKernel;
  cudaMalloc((void**)&cudaGaussKernel,kernelSize);

  /// Calculate gaussian kernel
  calculateGaussianKernel<<<DimGrid,DimBlock,kernelSize>>>(cudaGaussKernel, gaussianVariance, halfWidth);
  
  return(cudaGaussKernel);
}

